#define gamma_SEND(a,b,c)  gamma_send_flowctl(a,b,c)

#include <linux/gamma/libgamma.h>
#include <stdio.h>
#include <stdlib.h>
#include "measure.h"

#define MASTER_NODE 0
#define SLAVE_NODE 1
#define PORT 1
#define NOTIF_SEM 2
#define SYSCALL_TIME
#define PERFORMANCE
#undef PERF_SHORT
#undef DELAY
#undef FREQUENCIES


void handler_rec(void) {
	gamma_signal(NOTIF_SEM);
}


int main(int argc, char **argv) {

	unsigned char *data;
#ifdef PERFORMANCE
#ifdef SYSCALL_TIME
	double record_mc[NPROVE];
#endif
	double record_mm[NPROVE];
#endif
	int i,j,len;
	time_586
		t_send_s,
		t_rec;
#ifdef SYSCALL_TIME
	time_586
		t_send_e;
#endif
	double time_overhead;

	gamma_init(2,argc,argv);
	len = atoi(argv[1]);
	data = (unsigned char *)malloc(len*sizeof(unsigned char));

	/*  Pin communication data structures in RAM.  */
	gamma_mlock(data,len);

	/*  Init the message to be exchanged.  */
	for (j=0;j<len;j++) data[j] = (unsigned char)j;


	if (gamma_my_node() == MASTER_NODE) {

		/*  this is PING  */

		gamma_set_active_port(
			PORT,
			SLAVE_NODE,gamma_my_par_pid(),PORT,
			handler_rec,
			PORT,GO_BACK,
			data,len
		);

		/*  Computing the gamma_time() overhead...  */
		for (i=0;i<NPROVE;i++) {
			gamma_time(t_send_s);
 			gamma_time(t_rec);
			record_mm[i] = gamma_time_diff(t_rec, t_send_s);
		}
		{
		double Mmm,medmm,varmm;
		mean_variance(record_mm,&time_overhead,&Mmm,&medmm,&varmm);
		}

		gamma_sync();

		for (i=0;i<NPROVE;i++) {

#ifdef PERFORMANCE
			gamma_time(t_send_s);
#endif
			gamma_SEND(PORT,data,len);
#ifdef PERFORMANCE
#ifdef SYSCALL_TIME
			gamma_time(t_send_e);
#endif
#endif
			if ( gamma_wait(NOTIF_SEM,1) ) {
				printf("PING:  ERROR\n");
				fflush(stdout);
				gamma_exit();
				exit(-1);
			}
#ifdef PERFORMANCE
 			gamma_time(t_rec);
#endif

			/*
			 *  Force printout of kernel-level time measurements,
			 *  if enabled.  Otherwise nothing happens.
			 */
			gamma_timers(t_send_s,t_rec);

                        for (j=0;j<len;j++)
                                if (data[j] != (unsigned char)j) {
                                        int k;
                                        printf("PING:  data error\n");
                                        for (k=j-5;k<j+10;k++)
						printf("%d:%d ",k,data[k]);
                                        printf("\n");
                                        exit(0);
                                }

#ifdef PERFORMANCE
#ifdef SYSCALL_TIME
			record_mc[i] =
			gamma_time_diff(t_send_e, t_send_s)-time_overhead;
#endif
			record_mm[i] =
			gamma_time_diff(t_rec, t_send_s)/2-time_overhead;
#endif

		}

		gamma_exit();

#ifdef PERFORMANCE
#ifdef PERF_SHORT
		{
			double mmm,Mmm,medmm,varmm;
			mean_variance(record_mm,&mmm,&Mmm,&medmm,&varmm);
#ifdef DELAY
		        printf("%d %g\n",len,medmm);
#else
			if (!len)
			        printf("%d %g\n",len,medmm);
			else
			        printf("%d %g\n",len,len/medmm);
#endif
#ifdef FREQUENCIES
		        printf("variance: %g\n", varmm);
			frequencies(record_mm,mmm,Mmm);
#endif
		}
#else
		{
		printf("\n\n");
		printf("PERFORMANCE (microsec, MB/sec) (best < average < worst):\n\n");
		{
#ifdef SYSCALL_TIME
			double mmc,Mmc,medmc,varmc;
#endif
			double mmm,Mmm,medmm,varmm;
#ifdef SYSCALL_TIME
			mean_variance(record_mc,&mmc,&Mmc,&medmc,&varmc);
		        printf("sys call   %g < %g < %g\n", mmc,medmc,Mmc);
#endif
			mean_variance(record_mm,&mmm,&Mmm,&medmm,&varmm);
		        printf("delay       %g < %g < %g\n", mmm,medmm,Mmm);
		        printf("throughput  %g < %g < %g\n", len/mmm,len/medmm,len/Mmm);
			printf("\n\n");

		        printf("VARIANCE:  \n");
#ifdef SYSCALL_TIME
		        printf("sys call %g\n", varmc);
#endif
		        printf("delay %g\n", varmm);
			printf("\n\n");

#ifdef FREQUENCIES
		        printf("FREQUENCIES:\n");
#ifdef SYSCALL_TIME
			printf("sys call:\n");
			frequencies(record_mc,mmc,Mmc);
#endif
			printf("delay:\n");
			frequencies(record_mm,mmm,Mmm);
#endif
			fflush(stdout);

		}
		}
#endif
#endif
	}


	else {
		/*  this is PONG  */
		gamma_set_active_port(
			PORT,
			MASTER_NODE,gamma_my_par_pid(),PORT,
			handler_rec,
			PORT,GO_BACK,
			data,len
		);
		gamma_sync();
		for (i=0;i<NPROVE;i++) {
			if (!gamma_wait(NOTIF_SEM,1)) {
				gamma_SEND(PORT,data,len);
				continue;
			}
			printf("PONG:  ERROR\n");
			fflush(stdout);
			gamma_exit();
			exit(-1);
		}
		gamma_exit();
	}
return 0;
}
