/***************************************************************************
	NIC-independent macros for the GAMMA protocol
	Read COPYRIGHT file in this directory
****************************************************************************/


#define gmin(a,b) ((a)<=(b)?(a):(b))
#define gmax(a,b) ((a)>=(b)?(a):(b))

#ifdef MORE_INSTANCES_ON_SAME_NODE
#define DELAY_FOR_SMP busywait(50)
#else
#define DELAY_FOR_SMP
#endif

#define printg(a...) printk(KERN_EMERG a)

#ifdef DRIVER_DEBUG_EVERYTHING
#define DEBUG_E(a) {a;DELAY_FOR_SMP;}
#define DEBUG(a) {a;DELAY_FOR_SMP;}
#define DEBUG_F(a) {a;DELAY_FOR_SMP;}
#define DEBUG_L(a) {a;DELAY_FOR_SMP;}
#else
#ifdef DRIVER_DEBUG
#define DEBUG_E(a)
#define DEBUG(a) {a;DELAY_FOR_SMP;}
#define DEBUG_F(a) {a;DELAY_FOR_SMP;}
#define DEBUG_L(a) {a;DELAY_FOR_SMP;}
#else
#ifdef DRIVER_DEBUG_FLWCTL_RETX
#define DEBUG_E(a)
#define DEBUG(a)
#define DEBUG_F(a) {a;DELAY_FOR_SMP;}
#define DEBUG_L(a) {a;DELAY_FOR_SMP;}
#else
#ifdef DRIVER_DEBUG_LIGHT
#define DEBUG_E(a)
#define DEBUG(a)
#define DEBUG_F(a)
#define DEBUG_L(a) {a;DELAY_FOR_SMP;}
#else
#define DEBUG_E(a)
#define DEBUG(a)
#define DEBUG_F(a)
#define DEBUG_L(a)
#endif
#endif
#endif
#endif

#if CPU == 32

#define _set_gate(gate_addr,type,dpl,addr,seg) \
do { \
	int __d0, __d1; \
	__asm__ __volatile__ ("movw %%dx,%%ax\n\t" \
		"movw %4,%%dx\n\t" \
		"movl %%eax,%0\n\t" \
		"movl %%edx,%1" \
		:"=m" (*((long *) (gate_addr))), \
		"=m" (*(1+(long *) (gate_addr))), "=&a" (__d0), "=&d" (__d1) \
		:"i" ((short) (0x8000+(dpl<<13)+(type<<8))), \
		"3" ((char *) (addr)),"2" ((seg) << 16)); \
} while (0)

#define gamma_set_system_gate(n,funct) \
	_set_gate(idt_table+(n),15,3,funct,__KERNEL_CS);

#elif CPU == 64

#include <asm/desc.h>

#define gamma_set_system_gate(n,func) \
	_set_gate(&idt_table[n],GATE_INTERRUPT,(unsigned long)func,3,0);

#endif

/* the pte_offset_map() should be followed by a pte_unmap() if HIGHMEM in use */
#define user_space_to_phys(P,addr) \
(void *)virt_to_phys((void *)( \
    (unsigned long)page_address( \
        pte_page( \
            *pte_offset_map( \
                pmd_offset( \
                    pud_offset( \
                        pgd_offset( \
                            (P)->mm, \
                            (unsigned long)(addr) \
                        ), \
                        (unsigned long)(addr) \
                    ), \
                    (unsigned long)(addr) \
                ), \
                (unsigned long)(addr) \
            ) \
        ) \
    ) \
    | ((unsigned long)(addr) & 0x00000fffUL) \
))

/*  We should actually check the outcome of copy_***_user for segfaults...  */
#define gamma_copy_from_user(d,s,l) \
do { if (unlikely(copy_from_user(d,s,l))) break; } while (0)

#define gamma_copy_to_user(d,s,l) \
do { if (unlikely(copy_to_user(d,s,l))) break; } while (0)

#define recv_payload_1st_chunk(dest,src,skip,len) \
gamma_copy_to_user((void *)(dest),(void *)((u8 *)(src)+skip),len)

#define recv_payload_2nd_chunk(dest,src,skip1,skip2,len) \
gamma_copy_to_user((void *)(dest),(void *)((u8 *)(src)+skip1+skip2),len)

/*
 *  NOTE: most of what follows is NOT thread safe on SMP machines.
 *  Caller should raise spinlock.
 */

/*
 *  Shift user-space rec buffer's current ptr forward.
 *  Check for room availability.
 *  If not enough room in rec buffer, truncate the msg chunk
 *  by setting cnk_len equal to the available room.
 *  Decrement the available room by cnk_len.
 *  This macro is to prepare the download of a fresh packet into final dest.
 *  The fresh packet will be stored starting from address "rx_buffer".
 */
#define update_ptr_to_user_recv_buffer(par_pid,port,procnum,len,rx_buffer) \
{ \
rx_buffer = \
par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_progress; \
if (unlikely \
(par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_room < len))\
{ \
printg("rx_gamma(): message truncated!! %ld --> %u\n", \
len, par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_room); \
printg("par_pid %ld, procnum %ld, port %ld\n",par_pid,procnum,port); \
len = par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_room; \
}\
par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_room -= len; \
par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_progress = \
(void *)((u8 *) \
par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_progress \
	+ len); \
}

/*
 *  Get the pointer to current user-level recv buffer.
 */
#define get_ptr_to_user_recv_buffer(par_pid,port,procnum,rx_buffer) \
rx_buffer = \
par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_progress

/*  Set len of a just arrived msg.  */
#define set_msglen(par_pid,procnum,len) \
gamma_copy_to_user( \
	(u32 *)(par_proc_table[par_pid]->inst_table[procnum].msglen_ptr), \
	&(len), \
	sizeof(u32) \
)

#ifdef USE_DAMNICK
/*  Pre-notify a message arrival to the port semaphore.  */
#define pre_notify_arrival(par_pid,port,procnum,pre_notif_val) \
{ \
gamma_copy_from_user( \
	&pre_notif_val, \
	((unsigned long *) \
(par_proc_table[par_pid]->inst_table[procnum].pre_sem_table)+ \
par_proc_table[par_pid]->inst_table[procnum].port_table[port].semaphore), \
	sizeof(pre_notif_val) \
); \
pre_notif_val++; \
gamma_copy_to_user( \
	((unsigned long *) \
(par_proc_table[par_pid]->inst_table[procnum].pre_sem_table)+ \
par_proc_table[par_pid]->inst_table[procnum].port_table[port].semaphore), \
	&pre_notif_val, \
	sizeof(pre_notif_val) \
); \
}
#endif

#define reset_current_ptr(par_pid,port,procnum) \
{ \
par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_progress \
= par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_buffer; \
par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_room \
= par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_size; \
}

#define \
check_room_possibly_reset_current_ptr(par_pid,port,procnum) \
{ \
if (!(par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_room)) \
{ \
par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_progress \
= par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_buffer; \
par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_room \
= par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_size; \
} \
}

#define low_watermark_pt2pt(N) \
(LOW_WATERMARK < initial_credit[N] ? LOW_WATERMARK : initial_credit[N])

#define read_credit(node,credit) \
for (;;) { \
	int i; \
	if (node != BROADCAST) { \
		credit = atomic_read(&local_credit[node]); \
		break; \
	} \
	credit = 999999; \
	for (i=0; i<total_num_nodes; i++) \
		if (i != phys_whoami) { \
			if (atomic_read(&local_credit[i]) < (credit)) \
				credit = atomic_read(&local_credit[i]); \
			if (atomic_read(&local_credit[i]) < low_watermark_pt2pt(i)) { \
				credit = 0; \
				break; \
			} \
		} \
	break; \
}

#define dec_credit(node,amount) \
for (;;) { \
	int iii; \
	if (node != BROADCAST) { \
		atomic_sub(amount,&local_credit[node]); \
		break; \
	} \
	for (iii=0; iii<total_num_nodes; iii++) \
		if (iii != phys_whoami) \
			atomic_sub(amount,&local_credit[iii]); \
	break; \
}

#define refill_credit(done,node,credit,curr_credit) \
for (;;) { \
	read_credit(node,credit); \
	if (credit) { \
DEBUG_F( \
if (credit) { \
int i; \
printg("refill_credit():  credit refilled:\n"); \
for (i=0;i<total_num_nodes;i++) \
	printg("(%d) ", atomic_read(&local_credit[i])); \
printg(".\n"); \
}) \
		curr_credit = credit; \
		done = 1; \
		break; \
	} \
	done = 0; \
DEBUG_E({ \
int i; \
printg("refill_credit():  credit NOT refilled yet: "); \
for (i=0;i<total_num_nodes;i++) \
	printg("(%d) ", atomic_read(&local_credit[i])); \
printg("\n"); \
}) \
	break; \
}

#define incr_modulo(x,n) if (unlikely((x)+1 >= (n))) (x) = 0; else (x) = (x)+1;

#define wait_room_in_shared_ring(min,max,i) \
{ \
time_586 a,b; \
i = 1; \
gamma_time(a); \
for (;;) { \
	if (shared_ring_room(min,max) > SHARED_RING_LOW) { \
		break; \
	} \
	gamma_time(b); \
	if (gamma_time_diff_ticks(b,a) > (u64)MAX_CREDIT_WAIT*CLOCK) { \
		i = 0; \
		break; \
	} \
} \
}

#define busywait(n) \
{ \
	time_586 a,b; \
	gamma_time(a); \
	for (;;) { \
		gamma_time(b); \
		if (gamma_time_diff_ticks(b,a) > (u64)(n)*CLOCK) \
			break; \
	} \
}

#define tx_ring_is_full() (tx_ring_occupancy >= GAMMA_TX_RING_SIZE-5)

#define tx_ring_has_room() (tx_ring_occupancy < GAMMA_TX_RING_SIZE-20)

#define IRQ_enable_for_5usec(t_elaps,t_start) \
if (gamma_time_diff_ticks((t_elaps),(t_start)) > (unsigned long)9000*CLOCK) { \
	/* \
	 *  Open a 5 usec window for IRQs every 9 msec, necessary for \
	 *  the correct management of the OS clock (gettimeoday()). \
	 * \
	 */ \
	unsigned long flags; \
	local_save_flags(flags); \
	local_irq_enable(); \
	busywait(5); \
	local_irq_restore(flags); \
	gamma_time(t_start); \
}
