/***************************************************************************
	Read COPYRIGHT file in this directory
****************************************************************************/

int
sys_gamma_connect_port
(
	unsigned long par_pid_and_port_and_procnum,
	unsigned long partner_instnum,
	unsigned long partner_port_and_partner_par_pid,
	unsigned long semaphore,
	unsigned short buffer_kind_and_boundary_1p
)
{

	unsigned long par_pid,port,procnum,i;
#ifdef MORE_INSTANCES_ON_SAME_NODE
	int n,nrecv;
#endif
	port_entry *p;


	procnum = (unsigned char)par_pid_and_port_and_procnum;
	port = (unsigned short)(par_pid_and_port_and_procnum >> 8);
	par_pid = (unsigned char)(par_pid_and_port_and_procnum >> 24);

	p = &(par_proc_table[par_pid]->inst_table[procnum].port_table[port]);

	p->partner_par_pid = (unsigned char)partner_port_and_partner_par_pid;
	p->partner_instnum = partner_instnum;
	p->partner_port = (unsigned short)
		(partner_port_and_partner_par_pid >> 8);

	/*
	 *  Compute the phys node # and the "procnum"
	 *  of partner proc instance, if needed
	 */
	if (p->partner_par_pid != DONT_MIND) {
		if (partner_instnum == BROADCAST) {
			p->phys_partner_node = BROADCAST;
			p->partner_procnum = 0;
		}
		else {
			p->phys_partner_node =
				par_proc_table[p->partner_par_pid]->vm
				.node_list[partner_instnum];
			p->partner_procnum = 0;
			for (i=0;i<partner_instnum;i++)
				if (
					par_proc_table[p->partner_par_pid]->vm
					.node_list[i]
					==
					p->phys_partner_node
				)
					p->partner_procnum++;
		}
	}

#ifdef MORE_NICS_SAMENODE_ROUTING
	p->nic_idx = (p->phys_partner_node == BROADCAST) ?
		BCAST_ALL_NICS : routing_table[p->phys_partner_node];
#endif

        p->task = current;
        p->pid = current->pid;
	p->semaphore = semaphore;
	p->buffer_kind = buffer_kind_and_boundary_1p >> 8;
	p->boundary_1p = buffer_kind_and_boundary_1p & (unsigned short)0x00ff;
#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (p->phys_partner_node != BROADCAST)
		p->partner_is_intra_node =
			(p->phys_partner_node == (unsigned char)phys_whoami);
	else {
/*  Bcast: Check wether all receivers are on the local node or not.  */
/*
 *  Two cases: receivers are in the same group of the sender, or not.
 *  In the former case, no. of receivers = no. of group members - 1.
 *  Otherwise, no. of receivers = no. of group members.
 */
		n = par_proc_table[p->partner_par_pid]->vm.num_instances;
		if (par_pid == p->partner_par_pid)
			nrecv = n - 1;
		else
			nrecv = n;
		for (i=0;i<MAX_NUM_LOCAL_INST;i++) {
			if (
				par_proc_table[p->partner_par_pid]
				->inst_table[i].whoami
				< 0
			)
				break;
			if (
				par_proc_table[p->partner_par_pid]
				->inst_table[i].task
				== current
			)
				continue;
			nrecv--;
		}
		p->partner_is_intra_node = (!nrecv);
	}
	p->bcast_was_susp = 0;
#endif
	p->suspended_isend_handle = -GAMMA_ISEND_IS_DONE;
	p->current_tx_data1 = NULL;
	p->current_tx_len1 = 0;
	p->current_tx_data2 = NULL;
	p->current_tx_len2 = 0;
	p->current_rx_msglen = 0;

DEBUG(printg(
"sys_gamma_connect_port():"
"\tconnect par_pid %ld port %ld to par_pid %d inst %d port %d"
"\t(local node %ld local procnum %ld remote_node %d remote_procnum %d),"
"\tsemaphore %d,"
"\tbuffer kind %d,"
"\t1p boundary %d\n",
par_pid,port,
p->partner_par_pid, p->partner_instnum, p->partner_port,
phys_whoami,procnum, p->phys_partner_node, p->partner_procnum,
p->semaphore,
p->buffer_kind,
p->boundary_1p
));
#ifdef MORE_NICS_SAMENODE_ROUTING
DEBUG(printg(
	     "sys_gamma_connect_port():  precomputing NIC %d for remote node %d\n",
	     p->nic_idx,p->phys_partner_node ));
#endif

#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (!(p->partner_is_intra_node))
#endif
/*
 *  Prepare the fixed fields of the header,
 *  and set the sequence frame number (p->header[5].ss[1]) to zero.
 */
#ifdef MORE_NICS_SAMENODE_ROUTING
		make_send_header(
			p->phys_partner_node,
			p->partner_par_pid,
			p->partner_procnum,
			p->partner_port,ETH_GAMMA,0,0,0,p->header,
/* If broadcast, the computed sender field in the header is "wrong" in every subnet 
 * except the first. This should not be a problem. */
			(p->nic_idx == BCAST_ALL_NICS)? 0 : p->nic_idx
		);
#else
		make_send_header(
			p->phys_partner_node,
			p->partner_par_pid,
			p->partner_procnum,
			p->partner_port,ETH_GAMMA,0,0,0,p->header,0
		);
#endif

        return 0; 

}


int
sys_gamma_post_recv
(
	unsigned long par_pid,
	unsigned long procnum,
	unsigned long port,
	void *in_buffer,
	unsigned long in_size
)
{
	port_entry *p;

	p = &(par_proc_table[par_pid]->inst_table[procnum].port_table[port]);
/*  Added this to support Lars' GAMMA sockets in multithreaded programs  */
	p->task = current;
	p->pid = current->pid;
	p->in_buffer = in_buffer;
	p->in_size = in_size;
	p->in_progress = in_buffer;
	p->in_room = in_size;
	return 0;
}


int
sys_gamma_switch_to_buffer
(
	unsigned long par_pid,
	unsigned long procnum,
	unsigned long port,
	void *in_buffer,
	unsigned long in_size
)
{
	port_entry *p;

	p = &(par_proc_table[par_pid]->inst_table[procnum].port_table[port]);
	p->in_progress = in_buffer;
	p->in_room = in_size;
	return 0;
}


void *
sys_gamma_recvptr
(
	unsigned long par_pid,
	unsigned long procnum,
	unsigned long port
)
{
	port_entry *p;

	p = &(par_proc_table[par_pid]->inst_table[procnum].port_table[port]);
	return p->in_progress;
}
